# Copyright (C) 2019-2020 Intel Corporation
#
# SPDX-License-Identifier: MIT


class DatumaroPath:
    IMAGES_DIR = "images"
    ANNOTATIONS_DIR = "annotations"
    PCD_DIR = "point_clouds"
    RELATED_IMAGES_DIR = "related_images"
    MASKS_DIR = "masks"

    IMAGE_EXT = ".jpg"
    MASK_EXT = ".png"
