# Copyright (C) 2019-2020 Intel Corporation
#
# SPDX-License-Identifier: MIT


class CvatPath:
    IMAGES_DIR = "images"

    IMAGE_EXT = ".jpg"

    BUILTIN_ATTRS = {"occluded", "outside", "keyframe", "track_id"}
