# Copyright (C) 2019-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT


class YoloPath:
    DEFAULT_SUBSET_NAME = "train"
    SUBSET_NAMES = ["train", "valid"]
    RESERVED_CONFIG_KEYS = ["backup", "classes", "names"]
